@@ -8,7 +8,12 @@ import numpy as np
 import gym
 from tqdm.auto import tqdm
 
-from .preprocessing import create_scaler, Scaler
+from .preprocessing import (
+    create_scaler,
+    Scaler,
+    create_action_scaler,
+    ActionScaler,
+)
 from .augmentation import create_augmentation, AugmentationPipeline
 from .augmentation import DrQPipeline
 from .dataset import Episode, MDPDataset, Transition, TransitionMiniBatch
@@ -19,8 +24,16 @@ from .gpu import Device
 from .models.optimizers import OptimizerFactory
 from .models.encoders import EncoderFactory, create_encoder_factory
 from .models.q_functions import QFunctionFactory, create_q_func_factory
-from .argument_utility import check_scaler, ScalerArg, UseGPUArg
+from .argument_utility import (
+    check_scaler,
+    ScalerArg,
+    check_action_scaler,
+    ActionScalerArg,
+    UseGPUArg,
+)
 from .online.utility import get_action_size_from_env
+from .iterators import RoundIterator
+from .constants import IMPL_NOT_INITIALIZED_ERROR
 
 
 class ImplBase(metaclass=ABCMeta):
@@ -47,7 +60,9 @@ def _serialize_params(params: Dict[str, Any]) -> Dict[str, Any]:
     for key, value in params.items():
         if isinstance(value, Device):
             params[key] = value.get_id()
-        elif isinstance(value, (Scaler, EncoderFactory, QFunctionFactory)):
+        elif isinstance(
+            value, (Scaler, ActionScaler, EncoderFactory, QFunctionFactory)
+        ):
             params[key] = {
                 "type": value.get_type(),
                 "params": value.get_params(),
@@ -72,6 +87,11 @@ def _deseriealize_params(params: Dict[str, Any]) -> Dict[str, Any]:
             scaler_params = params["scaler"]["params"]
             scaler = create_scaler(scaler_type, **scaler_params)
             params[key] = scaler
+        if key == "action_scaler" and params["action_scaler"]:
+            scaler_type = params["action_scaler"]["type"]
+            scaler_params = params["action_scaler"]["params"]
+            action_scaler = create_action_scaler(scaler_type, **scaler_params)
+            params[key] = action_scaler
         elif key == "augmentation" and params["augmentation"]:
             augmentations = []
             for param in params[key]["augmentations"]:
@@ -100,6 +120,7 @@ class LearnableBase:
     _n_steps: int
     _gamma: float
     _scaler: Optional[Scaler]
+    _action_scaler: Optional[ActionScaler]
     _impl: Optional[ImplBase]
     _eval_results: DefaultDict[str, List[float]]
     _loss_history: DefaultDict[str, List[float]]
@@ -112,12 +133,14 @@ class LearnableBase:
         n_steps: int,
         gamma: float,
         scaler: ScalerArg,
+        action_scaler: ActionScalerArg,
     ):
         self._batch_size = batch_size
         self._n_frames = n_frames
         self._n_steps = n_steps
         self._gamma = gamma
         self._scaler = check_scaler(scaler)
+        self._action_scaler = check_action_scaler(action_scaler)
 
         self._impl = None
         self._eval_results = defaultdict(list)
@@ -272,7 +295,7 @@ class LearnableBase:
             fname: destination file path.
 
         """
-        assert self._impl is not None
+        assert self._impl is not None, IMPL_NOT_INITIALIZED_ERROR
         self._impl.save_model(fname)
 
     def load_model(self, fname: str) -> None:
@@ -286,7 +309,7 @@ class LearnableBase:
             fname: source file path.
 
         """
-        assert self._impl is not None
+        assert self._impl is not None, IMPL_NOT_INITIALIZED_ERROR
         self._impl.load_model(fname)
 
     def fit(
@@ -335,18 +358,28 @@ class LearnableBase:
 
         """
 
-        transitions = []
-        for episode in episodes:
-            transitions += episode.transitions
+        iterator = RoundIterator(
+            episodes,
+            batch_size=self._batch_size,
+            n_steps=self._n_steps,
+            gamma=self._gamma,
+            n_frames=self._n_frames,
+            shuffle=shuffle,
+        )
 
         # initialize scaler
         if self._scaler:
             self._scaler.fit(episodes)
 
+        # initialize action scaler
+        if self._action_scaler:
+            self._action_scaler.fit(episodes)
+
         # instantiate implementation
         if self._impl is None:
-            action_size = transitions[0].get_action_size()
-            observation_shape = tuple(transitions[0].get_observation_shape())
+            transition = iterator.transitions[0]
+            action_size = transition.get_action_size()
+            observation_shape = tuple(transition.get_observation_shape())
             self.create_impl(
                 self._process_observation_shape(observation_shape), action_size
             )
@@ -365,7 +398,7 @@ class LearnableBase:
         self._active_logger = logger
 
         # save hyperparameters
-        self._save_params(logger)
+        self.save_params(logger)
 
         # refresh evaluation metrics
         self._eval_results = defaultdict(list)
@@ -373,57 +406,38 @@ class LearnableBase:
         # refresh loss history
         self._loss_history = defaultdict(list)
 
-        # hold original dataset
-        env_transitions = transitions
-
         # training loop
         total_step = 0
         for epoch in range(n_epochs):
 
             # data augmentation
-            new_transitions = self._generate_new_data(env_transitions)
+            new_transitions = self._generate_new_data(iterator.transitions)
             if new_transitions:
-                transitions = env_transitions + new_transitions
-
-            # shuffle data
-            if shuffle:
-                indices = np.random.permutation(np.arange(len(transitions)))
-            else:
-                indices = np.arange(len(transitions))
+                iterator.set_ephemeral_transitions(new_transitions)
 
             # dict to add incremental mean losses to epoch
             epoch_loss = defaultdict(list)
 
-            n_iters = len(transitions) // self._batch_size
             range_gen = tqdm(
-                range(n_iters),
+                range(len(iterator)),
                 disable=not show_progress,
                 desc="Epoch %d" % int(epoch),
             )
 
+            iterator.reset()
+
             for itr in range_gen:
                 with logger.measure_time("step"):
                     # pick transitions
                     with logger.measure_time("sample_batch"):
-                        sampled_transitions = []
-                        head_index = itr * self.batch_size
-                        tail_index = head_index + self.batch_size
-                        for index in indices[head_index:tail_index]:
-                            sampled_transitions.append(transitions[index])
-
-                        batch = TransitionMiniBatch(
-                            transitions=sampled_transitions,
-                            n_frames=self._n_frames,
-                            n_steps=self._n_steps,
-                            gamma=self._gamma,
-                        )
+                        batch = next(iterator)
 
                     # update parameters
                     with logger.measure_time("algorithm_update"):
                         loss = self.update(epoch, total_step, batch)
 
                     # record metrics
-                    for name, val in zip(self._get_loss_labels(), loss):
+                    for name, val in zip(self.get_loss_labels(), loss):
                         if val is not None:
                             logger.add_metric(name, val)
                             epoch_loss[name].append(val)
@@ -440,7 +454,7 @@ class LearnableBase:
             # save loss to loss history dict
             self._loss_history["epoch"].append(epoch)
             self._loss_history["step"].append(total_step)
-            for name in self._get_loss_labels():
+            for name in self.get_loss_labels():
                 if name in epoch_loss:
                     self._loss_history[name].append(np.mean(epoch_loss[name]))
 
@@ -450,7 +464,7 @@ class LearnableBase:
             # save metrics
             logger.commit(epoch, total_step)
 
-            # save model parameters and greedy policy
+            # save model parameters
             if epoch % save_interval == 0:
                 logger.save_model(epoch, self)
 
@@ -540,7 +554,7 @@ class LearnableBase:
         """
         return None
 
-    def _get_loss_labels(self) -> List[str]:
+    def get_loss_labels(self) -> List[str]:
         raise NotImplementedError
 
     def _prepare_logger(
@@ -588,8 +602,14 @@ class LearnableBase:
             if test_score is not None:
                 self._eval_results[name].append(test_score)
 
-    def _save_params(self, logger: D3RLPyLogger) -> None:
-        assert self._impl
+    def save_params(self, logger: D3RLPyLogger) -> None:
+        """Saves configurations as params.json.
+
+        Args:
+            logger: logger object.
+
+        """
+        assert self._impl is not None, IMPL_NOT_INITIALIZED_ERROR
 
         # get hyperparameters without impl
         params = {}
@@ -683,6 +703,20 @@ class LearnableBase:
     def scaler(self, scaler: Scaler) -> None:
         self._scaler = scaler
 
+    @property
+    def action_scaler(self) -> Optional[ActionScaler]:
+        """Preprocessing action scaler.
+
+        Returns:
+            Optional[ActionScaler]: preprocessing action scaler.
+
+        """
+        return self._action_scaler
+
+    @action_scaler.setter
+    def action_scaler(self, action_scaler: ActionScaler) -> None:
+        self._action_scaler = action_scaler
+
     @property
     def impl(self) -> Optional[ImplBase]:
         """Implementation object.
